==> nomad_tools-0.2.0-py3-none-any.whl: nomad_tools/nomaddbjob.py <==
import logging
import queue
import threading
from typing import Callable, Dict, Iterable, List, Optional, TypeVar, Union

import requests

from . import nomadlib
from .common import json_loads, mynomad
from .nomadlib import Event, EventTopic, EventType

log = logging.getLogger(__name__)
T = TypeVar("T")


class NomadDbJob:
    """Represents relevant state cache from Nomad database of a single job"""

    def __init__(
        self,
        topics: List[str],
        select_event_cb: Callable[[Event], bool],
        init_cb: Callable[[], List[Event]],
        force_polling: Optional[bool] = None,
    ):
        """
        :param topic The topics to listen to, see Nomad event stream API documentation.
        :param select_event_cb: Filter only relevant events from Nomad event stream.
        :param init_cb Return a list of events to populate the database with and to poll. Has to be threadsafe!
        """
        self.topics: List[str] = topics
        self.init_cb: Callable[[], List[Event]] = init_cb
        self.select_event_cb: Callable[[Event], bool] = select_event_cb
        self.force_polling = force_polling
        self.queue: queue.Queue[Optional[List[Event]]] = queue.Queue()
        """Queue where database thread puts the received events"""
        self.job: Optional[nomadlib.Job] = None
        """Watched job. Is None if the job was deregistered."""
        self.jobversions: Dict[int, nomadlib.Job] = {}
        """Job version to job definition"""
        self.evaluations: Dict[str, nomadlib.Eval] = {}
        """Database of evaluations"""
        self.allocations: Dict[str, nomadlib.Alloc] = {}
        """Database of allocations"""
        self.deployments: Dict[str, nomadlib.Deploy] = {}
        """Database of deployments"""
        self.initialized = threading.Event()
        """Was init_cb called to initiliaze the database"""
        self.stopevent = threading.Event()
        """If set, the database thread should exit"""
        self.thread = threading.Thread(
            target=self.__thread_entry, name="db", daemon=True
        )
        assert self.topics
        assert not any(not x for x in topics)

    ###############################################################################

    def __thread_run_stream(self):
        log.debug(f"Starting listen Nomad stream with {' '.join(self.topics)}")
        with mynomad.stream(
            "event/stream",
            params={"topic": self.topics},
        ) as stream:
            for line in stream.iter_lines():
                if line:
                    data = json_loads(line)
                    events: List[Event] = [
                        Event(
                            EventTopic[event["Topic"]],
                            EventType[event["Type"]],
                            event["Payload"][event["Topic"]],
                            stream=True,
                        )
                        for event in data.get("Events", [])
                    ]
                    # log.debug(f"RECV EVENTS: {events}")
                    self.queue.put(events)
                if self.stopevent.is_set():
                    break

    def __thread_poll(self):
        """If listening to stream fails, we fallback to calling init_cb in a loop"""
        # Delay polling start up until init_cb has been called.
        self.initialized.wait()
        while not self.stopevent.wait(1):
            self.queue.put(self.init_cb())

    def __thread_entry(self):
        """Database thread entry"""
        try:
            try:
                if self.force_polling is True:
                    self.__thread_poll()
                else:
                    try:
                        self.__thread_run_stream()
                    except nomadlib.PermissionDenied as e:
                        if self.force_polling is False:
                            raise
                        else:
                            log.warning(
                                f"Falling to polling method because stream API returned permission denied: {e}"
                            )
                    self.__thread_poll()
            finally:
                log.debug("Nomad database thread exiting")
                self.queue.put(None)
        except requests.HTTPError as e:
            log.exception("http request failed")
            exit(1)

    ###############################################################################

    def start(self):
        """Start the database thread"""
        assert (
            mynomad.namespace
        ), "Nomad namespace has to be set before starting to listen"
        self.thread.start()

    def select_is_in_db(self, e: Event):
        """Select events that are already in the database"""
        if e.topic == EventTopic.Evaluation:
            return e.data["ID"] in self.evaluations
        elif e.topic == EventTopic.Allocation:
            return e.data["ID"] in self.allocations
        elif e.topic == EventTopic.Deployment:
            return e.data["ID"] in self.deployments
        elif e.topic == EventTopic.Job:
            # Get the first available job ID to compare.
            jobid = (
                self.job.ID
                if self.job is not None
                else next((job.ID for job in self.jobversions.values()), None)
            )
            # Explicitly filter against job ID. Also compare versions.
            return (jobid is None or e.data["ID"] == jobid) and e.data[
                "Version"
            ] in self.jobversions
        return False

    def _add_event_to_db(self, e: Event):
        """Update database state to reflect received event"""
        if e.topic == EventTopic.Job:
            job = nomadlib.Job(e.data)
            self.jobversions[job.get("Version")] = job
            if e.type == EventType.JobDeregistered:
                self.job = None
            elif self.job is None or job.ModifyIndex >= self.job.ModifyIndex:
                # self.job follows newest modify index.
                self.job = job
        elif e.topic == EventTopic.Evaluation:
            if e.type == EventType.JobDeregistered:
                self.job = None
            self.evaluations[e.data["ID"]] = nomadlib.Eval(e.data)
        elif e.topic == EventTopic.Allocation:
            # Events from event stream are missing JobVersion. Try to preserve it here by preserving keys.
            self.allocations[e.data["ID"]] = nomadlib.Alloc(
                {**self.allocations.get(e.data["ID"], {}), **e.data}
            )
        elif e.topic == EventTopic.Deployment:
            self.deployments[e.data["ID"]] = nomadlib.Deploy(e.data)

    def handle_event(self, e: Event) -> bool:
        if self._select_new_event(e):
            if self.select_is_in_db(e) or self.select_event_cb(e):
                # log.debug(f"EVENT: {e}")
                # log.debug(f"EVENT: {e} {e.data}")
                self._add_event_to_db(e)
                return True
            else:
                # log.debug(f"USER FILTERED: {e}")
                pass
        else:
            # log.debug(f"OLD EVENT: {e}")
            pass
        return False

    def handle_events(self, events: List[Event]) -> List[Event]:
        """From a list of events, filter out ignored and add the rest to database"""
        return [e for e in events if self.handle_event(e)]

    @staticmethod
    def apply_selects(
        e: Event,
        job_select: Callable[[nomadlib.Job], bool],
        eval_select: Callable[[nomadlib.Eval], bool],
        alloc_select: Callable[[nomadlib.Alloc], bool],
        deploy_select: Callable[[nomadlib.Deploy], bool],
    ) -> bool:
        """Apply specific selectors depending on event type"""
        return e.apply(job_select, eval_select, alloc_select, deploy_select)

    def _select_new_event(self, e: Event):
        """Select events which are newer than those in the database"""
        job_select: Callable[[nomadlib.Job], bool] = lambda _: True
        eval_select: Callable[[nomadlib.Eval], bool] = (
            lambda eval: eval.ID not in self.evaluations
            or eval.ModifyIndex > self.evaluations[eval.ID].ModifyIndex
        )
        alloc_select: Callable[[nomadlib.Alloc], bool] = (
            lambda alloc: alloc.ID not in self.allocations
            or alloc.ModifyIndex > self.allocations[alloc.ID].ModifyIndex
        )
        deploy_select: Callable[[nomadlib.Deploy], bool] = (
            lambda deploy: deploy.ID not in self.deployments
            or deploy.ModifyIndex > self.deployments[deploy.ID].ModifyIndex
        )
        return e.data["Namespace"] == mynomad.namespace and self.apply_selects(
            e, job_select, eval_select, alloc_select, deploy_select
        )

    def stop(self):
        log.debug("Stopping listen Nomad stream")
        self.initialized.set()
        self.stopevent.set()

    def join(self):
        # Not joining - neither requests nor stream API allow for timeouts.
        # self.thread.join()
        pass

    def events(self) -> Iterable[List[Event]]:
        """Nomad stream returns Events array. Iterate over batches of events returned from Nomad stream"""
        assert self.thread.is_alive(), "Thread not alive"
        if not self.initialized.is_set():
            events = self.init_cb()
            events = self.handle_events(events)
            yield events
            self.initialized.set()
        log.debug("Starting getting events from thread")
        while not self.queue.empty() or (
            self.thread.is_alive() and not self.stopevent.is_set()
        ):
            events = self.queue.get()
            if events is None:
                break
            yield self.handle_events(events)
        log.debug("db exiting")

    def get_allocation_jobmodifyindex(
        self, alloc: nomadlib.Alloc, default: T = None
    ) -> Union[T, int]:
        """Given an allocation, return associated JobModifyIndex"""
        if alloc.JobVersion is not None:
            job = self.jobversions.get(alloc.JobVersion)
            if job:
                return job.JobModifyIndex
        evaluation = self.evaluations.get(alloc.EvalID)
        if evaluation:
            if evaluation.JobModifyIndex is not None:
                return evaluation.JobModifyIndex
            if evaluation.DeploymentID is not None:
                deployment = self.deployments.get(evaluation.DeploymentID)
                if deployment:
                    return deployment.JobModifyIndex
        return default

    def get_allocation_jobversion(
        self, alloc: nomadlib.Alloc, default: T = None
    ) -> Union[T, int]:
        """Given an allocation return the job version associated with that allocation"""
        # If alloc JobVersion is missing, try to set it.
        if alloc.JobVersion is None:
            evaluation = self.evaluations.get(alloc.EvalID)
            if evaluation:
                # If we can find Job from JobModifyIndex, lets find it.
                if evaluation.JobModifyIndex is not None:
                    job = self.find_job_from_modifyindex(evaluation.JobModifyIndex)
                    if job is not None:
                        alloc.JobVersion = job.Version
                # Otherwise we may find deployment that may have jobversion.
                if alloc.JobVersion is None and evaluation.DeploymentID is not None:
                    deployment = self.deployments.get(evaluation.DeploymentID)
                    if deployment:
                        alloc.JobVersion = deployment.JobVersion
        return alloc.JobVersion if alloc.JobVersion is not None else default

    def find_job_from_modifyindex(self, jobmodifyindex: int) -> Optional[nomadlib.Job]:
        # Note that job versions may not be in JobModifyIndex order.
        # Job versions of previous job (after purge and start) may be here.
        # Sort explicitly over JobModifyIndex
        for job in sorted(
            self.jobversions.values(), key=lambda job: job.JobModifyIndex, reverse=True
        ):
            if job.JobModifyIndex <= jobmodifyindex:
                return job
        return None

    def find_jobversion_from_modifyindex(self, jobmodifyindex: int) -> Optional[int]:
        ret = self.find_job_from_modifyindex(jobmodifyindex)
        return ret.Version if ret else None

==> nomad_tools-0.2.1-py3-none-any.whl: nomad_tools/nomaddbjob.py <==
import logging
import queue
import threading
from typing import Callable, Dict, Iterable, List, Optional, TypeVar, Union

import requests

from . import flagdebug, nomadlib
from .common import eprint, json_loads, mynomad
from .nomadlib import Event, EventTopic, EventType

log = logging.getLogger(__name__)
T = TypeVar("T")


class NomadDbJob:
    """Represents relevant state cache from Nomad database of a single job"""

    def __init__(
        self,
        topics: List[str],
        select_event_cb: Callable[[Event], bool],
        init_cb: Callable[[], List[Event]],
        force_polling: Optional[bool] = None,
    ):
        """
        :param topic The topics to listen to, see Nomad event stream API documentation.
        :param select_event_cb: Filter only relevant events from Nomad event stream.
        :param init_cb Return a list of events to populate the database with and to poll. Has to be threadsafe!
        """
        self.topics: List[str] = topics
        self.init_cb: Callable[[], List[Event]] = init_cb
        self.select_event_cb: Callable[[Event], bool] = select_event_cb
        self.force_polling = force_polling
        self.queue: queue.Queue[Optional[List[Event]]] = queue.Queue()
        """Queue where database thread puts the received events"""
        self.job_deregistered: bool = False
        """Is set to true if received job deregistration event"""
        self.job: Optional[nomadlib.Job] = None
        """Watched job definition. Is not None means the job was at least once received."""
        self.jobversions: Dict[int, nomadlib.Job] = {}
        """Job version to job definition"""
        self.evaluations: Dict[str, nomadlib.Eval] = {}
        """Database of evaluations"""
        self.allocations: Dict[str, nomadlib.Alloc] = {}
        """Database of allocations"""
        self.deployments: Dict[str, nomadlib.Deploy] = {}
        """Database of deployments"""
        self.initialized = threading.Event()
        """Was init_cb called to initiliaze the database"""
        self.stopevent = threading.Event()
        """If set, the database thread should exit"""
        self.thread = threading.Thread(
            target=self.__thread_entry, name=self.__class__.__name__, daemon=True
        )
        assert self.topics
        assert not any(not x for x in topics)

    ###############################################################################

    def __thread_run_stream(self):
        log.debug(f"Starting listen Nomad stream with {' '.join(self.topics)}")
        with mynomad.stream(
            "event/stream",
            params={"topic": self.topics},
        ) as stream:
            for line in stream.iter_lines(decode_unicode=True):
                if line == "event dropped from buffer":
                    # https://github.com/hashicorp/nomad/blob/main/nomad/stream/event_buffer.go#L273
                    log.error(line)
                    continue
                if line:
                    data = json_loads(line)
                    events: List[Event] = [
                        Event(
                            EventTopic[event["Topic"]],
                            EventType[event["Type"]],
                            event["Payload"][event["Topic"]],
                            stream=True,
                        )
                        for event in data.get("Events", [])
                    ]
                    if flagdebug.debug("recv") == 1:
                        for e in events:
                            eprint(f"RECVEVENTS:{data['Index']} {e}")
                    elif flagdebug.debug("recv") >= 2:
                        for e in events:
                            eprint(f"RECVEVENTS:{data['Index']} {e} {e.data}")
                    self.queue.put(events)
                if self.stopevent.is_set():
                    break

    def __thread_poll(self):
        """If listening to stream fails, we fallback to calling init_cb in a loop"""
        # Delay polling start up until init_cb has been called.
        self.initialized.wait()
        while not self.stopevent.wait(1):
            self.queue.put(self.init_cb())

    def __thread_entry(self):
        """Database thread entry"""
        try:
            try:
                if self.force_polling is True:
                    self.__thread_poll()
                else:
                    try:
                        self.__thread_run_stream()
                    except nomadlib.PermissionDenied as e:
                        if self.force_polling is False:
                            raise
                        else:
                            log.warning(
                                f"Falling to polling method because stream API returned permission denied: {e}"
                            )
                    self.__thread_poll()
            finally:
                log.debug("Nomad database thread exiting")
                self.queue.put(None)
        except requests.HTTPError:
            log.exception("http request failed")
            exit(1)

    ###############################################################################

    def start(self):
        """Start the database thread"""
        assert (
            mynomad.namespace
        ), "Nomad namespace has to be set before starting to listen"
        self.thread.start()

    def select_is_in_db(self, e: Event):
        """Select events that are already in the database"""
        if e.topic == EventTopic.Evaluation:
            return e.data["ID"] in self.evaluations
        elif e.topic == EventTopic.Allocation:
            return e.data["ID"] in self.allocations
        elif e.topic == EventTopic.Deployment:
            return e.data["ID"] in self.deployments
        elif e.topic == EventTopic.Job:
            # Get the first available job ID to compare.
            jobid = (
                self.job.ID
                if self.job is not None
                else next((job.ID for job in self.jobversions.values()), None)
            )
            # Explicitly filter against job ID. Also compare versions.
            return (jobid is None or e.data["ID"] == jobid) and e.data[
                "Version"
            ] in self.jobversions
        return False

    def _add_event_to_db(self, e: Event):
        """Update database state to reflect received event"""
        actionstr = ""
        if e.topic == EventTopic.Job:
            job = e.job()
            self.jobversions[job.get("Version")] = job
            if self.job is None or job.ModifyIndex >= self.job.ModifyIndex:
                # self.job follows newest modify index.
                self.job = job
                if e.type == EventType.JobDeregistered:
                    if not self.job_deregistered:
                        self.job_deregistered = True
                        actionstr = "JobDeregistered because job"
                else:
                    if self.job_deregistered:
                        self.job_deregistered = False
                        actionstr = "JobRegister"
        elif e.topic == EventTopic.Evaluation:
            # Handle job deregister event by clearing self.job.
            eval = e.eval()
            if (
                self.job
                and eval.Status == "complete"
                and (
                    e.type == EventType.JobDeregistered
                    or (
                        e.type == EventType.EvaluationUpdated
                        and eval.TriggeredBy == "job-deregister"
                    )
                )
            ):
                actionstr = "JobDeregister because eval"
                self.job_deregistered = True
            self.evaluations[e.data["ID"]] = eval
        elif e.topic == EventTopic.Allocation:
            # Events from event stream are missing JobVersion. Try to preserve it here by preserving keys.
            self.allocations[e.data["ID"]] = nomadlib.Alloc(
                {**self.allocations.get(e.data["ID"], {}), **e.data}
            )
        elif e.topic == EventTopic.Deployment:
            self.deployments[e.data["ID"]] = e.deployment()
        return actionstr

    def handle_event(self, e: Event) -> bool:
        if self._select_new_event(e):
            if self.select_is_in_db(e) or self.select_event_cb(e):
                actionstr = self._add_event_to_db(e)
                if flagdebug.debug("events") == 1:
                    eprint(f"EVENT: {e} {actionstr}")
                if flagdebug.debug("events") >= 2:
                    eprint(f"EVENT: {e} {e.data} {actionstr}")
                return True
            else:
                if flagdebug.debug("db"):
                    log.debug(f"EVENTFILTERED: {e}")
                pass
        else:
            if flagdebug.debug("db"):
                log.debug(f"OLDEVENT: {e}")
            pass
        return False

    def handle_events(self, events: List[Event]) -> List[Event]:
        """From a list of events, filter out ignored and add the rest to database"""
        return [e for e in events if self.handle_event(e)]

    @staticmethod
    def apply_selects(
        e: Event,
        job_select: Callable[[nomadlib.Job], bool],
        eval_select: Callable[[nomadlib.Eval], bool],
        alloc_select: Callable[[nomadlib.Alloc], bool],
        deploy_select: Callable[[nomadlib.Deploy], bool],
    ) -> bool:
        """Apply specific selectors depending on event type"""
        return e.apply(job_select, eval_select, alloc_select, deploy_select)

    def _select_new_event(self, e: Event):
        """Select events which are newer than those in the database"""
        job_select: Callable[[nomadlib.Job], bool] = lambda _: True
        eval_select: Callable[[nomadlib.Eval], bool] = (
            lambda eval: eval.ID not in self.evaluations
            or eval.ModifyIndex > self.evaluations[eval.ID].ModifyIndex
        )
        alloc_select: Callable[[nomadlib.Alloc], bool] = (
            lambda alloc: alloc.ID not in self.allocations
            or alloc.ModifyIndex > self.allocations[alloc.ID].ModifyIndex
        )
        deploy_select: Callable[[nomadlib.Deploy], bool] = (
            lambda deploy: deploy.ID not in self.deployments
            or deploy.ModifyIndex > self.deployments[deploy.ID].ModifyIndex
        )
        return e.data["Namespace"] == mynomad.namespace and self.apply_selects(
            e, job_select, eval_select, alloc_select, deploy_select
        )

    def stop(self):
        log.debug("Stopping listen Nomad stream")
        self.initialized.set()
        self.stopevent.set()

    def join(self):
        # Not joining - neither requests nor stream API allow for timeouts.
        # self.thread.join()
        pass

    def events(self) -> Iterable[List[Event]]:
        """Nomad stream returns Events array. Iterate over batches of events returned from Nomad stream"""
        assert self.thread.is_alive(), "Thread not alive"
        if not self.initialized.is_set():
            events = self.init_cb()
            events = self.handle_events(events)
            yield events
            self.initialized.set()
        log.debug("Starting getting events from thread")
        while not self.queue.empty() or (
            self.thread.is_alive() and not self.stopevent.is_set()
        ):
            events = self.queue.get()
            if events is None:
                break
            yield self.handle_events(events)
        log.debug("db exiting")

    def get_allocation_jobmodifyindex(
        self, alloc: nomadlib.Alloc, default: T = None
    ) -> Union[T, int]:
        """Given an allocation, return associated JobModifyIndex"""
        if alloc.JobVersion is not None:
            job = self.jobversions.get(alloc.JobVersion)
            if job:
                return job.JobModifyIndex
        evaluation = self.evaluations.get(alloc.EvalID)
        if evaluation:
            if evaluation.JobModifyIndex is not None:
                return evaluation.JobModifyIndex
            if evaluation.DeploymentID is not None:
                deployment = self.deployments.get(evaluation.DeploymentID)
                if deployment:
                    return deployment.JobModifyIndex
        return default

    def get_allocation_jobversion(
        self, alloc: nomadlib.Alloc, default: T = None
    ) -> Union[T, int]:
        """Given an allocation return the job version associated with that allocation"""
        # If alloc JobVersion is missing, try to set it.
        if alloc.JobVersion is None:
            evaluation = self.evaluations.get(alloc.EvalID)
            if evaluation:
                # If we can find Job from JobModifyIndex, lets find it.
                if evaluation.JobModifyIndex is not None:
                    job = self.find_job_from_modifyindex(evaluation.JobModifyIndex)
                    if job is not None:
                        alloc.JobVersion = job.Version
                # Otherwise we may find deployment that may have jobversion.
                if alloc.JobVersion is None and evaluation.DeploymentID is not None:
                    deployment = self.deployments.get(evaluation.DeploymentID)
                    if deployment:
                        alloc.JobVersion = deployment.JobVersion
        return alloc.JobVersion if alloc.JobVersion is not None else default

    def find_job_from_modifyindex(self, jobmodifyindex: int) -> Optional[nomadlib.Job]:
        # Note that job versions may not be in JobModifyIndex order.
        # Job versions of previous job (after purge and start) may be here.
        # Sort explicitly over JobModifyIndex
        for job in sorted(
            self.jobversions.values(), key=lambda job: job.JobModifyIndex, reverse=True
        ):
            if job.JobModifyIndex <= jobmodifyindex:
                return job
        return None

    def find_jobversion_from_modifyindex(self, jobmodifyindex: int) -> Optional[int]:
        ret = self.find_job_from_modifyindex(jobmodifyindex)
        return ret.Version if ret else None

    def seen_job(self) -> bool:
        return self.job is not None
